# -*- coding: utf-8; mode: python; -*-
#
#  GStreamer Debug Viewer - View and analyze GStreamer debug log files
#
#  Copyright (C) 2007 René Stadler <mail@renestadler.de>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 3 of the License, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  this program.  If not, see <http://www.gnu.org/licenses/>.

"""GStreamer Debug Viewer timeline widget plugin."""

import logging

from GstDebugViewer import Common, Data
from GstDebugViewer.GUI.colors import LevelColorThemeTango, ThreadColorThemeTango
from GstDebugViewer.Plugins import FeatureBase, PluginBase

from gettext import gettext as _
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk
import cairo


def iter_model_reversed(model):

    count = model.iter_n_children(None)
    for i in range(count - 1, 0, -1):
        yield model[i]


class LineFrequencySentinel (object):

    def __init__(self, model):

        self.model = model
        self.clear()

    def clear(self):

        self.data = None
        self.n_partitions = None
        self.partitions = None
        self.step = None
        self.ts_range = None

    def _search_ts(self, target_ts, first_index, last_index):

        model_get = self.model.get_value
        model_iter_nth_child = self.model.iter_nth_child
        col_id = self.model.COL_TIME

        # TODO: Rewrite using a lightweight view object + bisect.

        while True:
            middle = (last_index - first_index) // 2 + first_index
            if middle == first_index:
                return first_index
            ts = model_get(model_iter_nth_child(None, middle), col_id)
            if ts < target_ts:
                first_index = middle
            elif ts > target_ts:
                last_index = middle
            else:
                return middle

    def run_for(self, n):

        if n == 0:
            raise ValueError("illegal value for n")

        self.n_partitions = n

    def process(self):

        model = self.model
        result = []
        partitions = []

        first_ts = None
        for row in self.model:
            first_ts = row[model.COL_TIME]
            if first_ts is not None:
                break

        if first_ts is None:
            return

        last_ts = None
        i = 0
        UNPARSABLE_LIMIT = 500
        for row in iter_model_reversed(self.model):
            last_ts = row[model.COL_TIME]
            # FIXME: We ignore 0 here (unparsable lines!), this should be
            # handled differently!
            i += 1
            if i == UNPARSABLE_LIMIT:
                break
            if last_ts:
                last_index = row.path[0]
                break

        if last_ts is None or last_ts < first_ts:
            return

        step = int(float(last_ts - first_ts) / float(self.n_partitions))

        YIELD_LIMIT = 100
        limit = YIELD_LIMIT

        first_index = 0
        target_ts = first_ts + step
        old_found = 0
        while target_ts < last_ts:
            limit -= 1
            if limit == 0:
                limit = YIELD_LIMIT
                yield True
            found = self._search_ts(target_ts, first_index, last_index)
            result.append(found - old_found)
            partitions.append(found)
            old_found = found
            first_index = found
            target_ts += step

        if step == 0:
            result = []
            partitions = []

        self.step = step
        self.data = result
        self.partitions = partitions
        self.ts_range = (first_ts, last_ts,)


class LevelDistributionSentinel (object):

    def __init__(self, freq_sentinel, model):

        self.freq_sentinel = freq_sentinel
        self.model = model
        self.data = []

    def clear(self):

        del self.data[:]

    def process(self):

        MAX_LEVELS = 9
        YIELD_LIMIT = 10000
        y = YIELD_LIMIT

        model_get = self.model.get_value
        model_next = self.model.iter_next
        id_time = self.model.COL_TIME
        id_level = self.model.COL_LEVEL
        del self.data[:]
        data = self.data
        i = 0
        partitions_i = 0
        partitions = self.freq_sentinel.partitions
        counts = [0] * MAX_LEVELS
        tree_iter = self.model.get_iter_first()

        if not partitions:
            return

        level_index = 0
        level_iter = None

        finished = False
        while tree_iter:
            y -= 1
            if y == 0:
                y = YIELD_LIMIT
                yield True
            if level_iter is None:
                stop_index = level_index + 512
                levels = self.model.get_value_range(id_level,
                                                    level_index, stop_index)
                level_index = stop_index
                level_iter = iter(levels)
            try:
                level = level_iter.__next__()
            except StopIteration:
                level_iter = None
                continue
            while i > partitions[partitions_i]:
                data.append(tuple(counts))
                counts = [0] * MAX_LEVELS
                partitions_i += 1
                if partitions_i == len(partitions):
                    finished = True
                    break
            if finished:
                break
            counts[level] += 1
            i += 1

        # Now handle the last one:
        data.append(tuple(counts))

        yield False


class UpdateProcess (object):

    def __init__(self, freq_sentinel, dist_sentinel):

        self.freq_sentinel = freq_sentinel
        self.dist_sentinel = dist_sentinel
        self.is_running = False
        self.dispatcher = Common.Data.GSourceDispatcher()

    def __process(self):

        if self.freq_sentinel is None or self.dist_sentinel is None:
            return

        self.is_running = True

        for x in self.freq_sentinel.process():
            yield True

        self.handle_sentinel_finished(self.freq_sentinel)

        for x in self.dist_sentinel.process():
            yield True
            self.handle_sentinel_progress(self.dist_sentinel)

        self.is_running = False

        self.handle_sentinel_finished(self.dist_sentinel)
        self.handle_process_finished()

        yield False

    def run(self):

        if self.is_running:
            return

        self.dispatcher(self.__process())

    def abort(self):

        if not self.is_running:
            return

        self.dispatcher.cancel()
        self.is_running = False

    def handle_sentinel_progress(self, sentinel):

        pass

    def handle_sentinel_finished(self, sentinel):

        pass

    def handle_process_finished(self):

        pass


class VerticalTimelineWidget (Gtk.DrawingArea):

    __gtype_name__ = "GstDebugViewerVerticalTimelineWidget"

    def __init__(self, log_view):

        GObject.GObject.__init__(self)

        self.logger = logging.getLogger("ui.vtimeline")

        self.log_view = log_view
        self.theme = ThreadColorThemeTango()
        self.params = None
        self.thread_colors = {}
        self.next_thread_color = 0

        try:
            self.set_tooltip_text(_("Vertical timeline\n"
                                    "Different colors represent different threads"))
        except AttributeError:
            # Compatibility.
            pass

    def do_draw(self, ctx):

        alloc = self.get_allocation()
        x = alloc.x
        y = alloc.y
        w = alloc.width
        h = alloc.height

        # White background rectangle.
        ctx.set_line_width(0.)
        ctx.rectangle(0, 0, w, h)
        ctx.set_source_rgb(1., 1., 1.)
        ctx.fill()
        ctx.new_path()

        if self.params is None:
            self.__update_params()

        if self.params is None:
            return

        first_y, cell_height, data = self.params
        if len(data) < 2:
            return
        first_ts, last_ts = data[0][0], data[-1][0]
        ts_range = last_ts - first_ts
        if ts_range == 0:
            return

        ctx.set_line_width(1.)
        ctx.set_source_rgb(0., 0., 0.)

        half_height = cell_height // 2 - .5
        quarter_height = cell_height // 4 - .5
        first_y += half_height
        for i, i_data in enumerate(data):
            ts, thread = i_data
            if thread in self.thread_colors:
                ctx.set_source_rgb(*self.thread_colors[thread])
            else:
                self.next_thread_color += 1
                if self.next_thread_color == len(self.theme.colors):
                    self.next_thread_color = 0
                color = self.theme.colors[
                    self.next_thread_color][0].float_tuple()
                self.thread_colors[thread] = color
                ctx.set_source_rgb(*color)
            ts_fraction = float(ts - first_ts) / ts_range
            ts_offset = ts_fraction * h
            row_offset = first_y + i * cell_height
            ctx.move_to(-.5, ts_offset)
            ctx.line_to(half_height, ts_offset)
            ctx.line_to(w - quarter_height, row_offset)
            ctx.stroke()
            ctx.line_to(w - quarter_height, row_offset)
            ctx.line_to(w + .5, row_offset - half_height)
            ctx.line_to(w + .5, row_offset + half_height)
            ctx.fill()
        return True

    def do_configure_event(self, event):

        self.params = None
        self.queue_draw()

        return False

    def do_get_preferred_width(self):

        return 64, 64  # FIXME

    def clear(self):

        self.params = None
        self.thread_colors.clear()
        self.next_thread_color = 0
        self.queue_draw()

    def __update_params(self):

        # FIXME: Ideally we should take the vertical position difference of the
        # view into account (which is 0 with the current UI layout).

        view = self.log_view
        model = view.get_model()
        visible_range = view.get_visible_range()
        if visible_range is None:
            return
        start_path, end_path = visible_range

        if not start_path or not end_path:
            return

        column = view.get_column(0)
        bg_rect = view.get_background_area(start_path, column)
        cell_height = bg_rect.height
        cell_rect = view.get_cell_area(start_path, column)
        try:
            first_y = view.convert_bin_window_to_widget_coords(
                cell_rect.x, cell_rect.y)[1]
        except (AttributeError, SystemError,):
            # AttributeError is with PyGTK before 2.12.  SystemError is raised
            # with PyGTK 2.12.0, pygtk bug #479012.
            first_y = cell_rect.y % cell_height

            global _warn_tree_view_coords
            try:
                _warn_tree_view_coords
            except NameError:
                self.logger.warning("tree view coordinate conversion method "
                                    "not available, using aproximate offset")
                # Only warn once:
                _warn_tree_view_coords = True

        data = []
        tree_iter = model.get_iter(start_path)
        if tree_iter is None:
            return
        while model.get_path(tree_iter) != end_path:
            data.append(
                model.get(tree_iter, model.COL_TIME, model.COL_THREAD))
            tree_iter = model.iter_next(tree_iter)

        self.params = (first_y, cell_height, data,)

    def update(self):

        self.params = None
        self.queue_draw()


class TimelineWidget (Gtk.DrawingArea):

    __gtype_name__ = "GstDebugViewerTimelineWidget"

    __gsignals__ = {"change-position": (GObject.SignalFlags.RUN_LAST,
                                        None,
                                        (GObject.TYPE_INT,),)}

    def __init__(self):

        GObject.GObject.__init__(self)

        self.logger = logging.getLogger("ui.timeline")

        self.add_events(Gdk.EventMask.BUTTON1_MOTION_MASK |
                        Gdk.EventMask.BUTTON_PRESS_MASK |
                        Gdk.EventMask.BUTTON_RELEASE_MASK)

        self.process = UpdateProcess(None, None)
        self.process.handle_sentinel_progress = self.__handle_sentinel_progress
        self.process.handle_sentinel_finished = self.__handle_sentinel_finished

        self.model = None
        self.__offscreen = None
        self.__offscreen_size = (0, 0)
        self.__offscreen_dirty = (0, 0)

        self.__position_ts_range = None

        try:
            self.set_tooltip_text(_("Log event histogram\n"
                                    "Different colors represent different log-levels"))
        except AttributeError:
            # Compatibility.
            pass

    def __handle_sentinel_progress(self, sentinel):

        if sentinel == self.process.dist_sentinel:
            old_progress = self.__dist_sentinel_progress
            new_progress = len(sentinel.data)
            if new_progress - old_progress >= 32:
                self.__invalidate_offscreen(old_progress, new_progress)
                self.__dist_sentinel_progress = new_progress

    def __handle_sentinel_finished(self, sentinel):

        if sentinel == self.process.freq_sentinel:
            self.__invalidate_offscreen(0, -1)
        else:
            self.__invalidate_offscreen(self.__dist_sentinel_progress, -1)

    def __ensure_offscreen(self):

        alloc = self.get_allocation()
        if self.__offscreen_size == (alloc.width, alloc.height):
            return

        self.__offscreen = cairo.ImageSurface(
            cairo.FORMAT_ARGB32, alloc.width, alloc.height)
        self.__offscreen_size = (alloc.width, alloc.height)
        self.__offscreen_dirty = (0, alloc.width)
        if not self.__offscreen:
            self.__offscreen_size = (0, 0)
            raise ValueError("could not obtain offscreen image surface")

    def __invalidate_offscreen(self, start, stop):

        alloc = self.get_allocation()
        if stop < 0:
            stop += alloc.width

        dirty_start, dirty_stop = self.__offscreen_dirty
        if dirty_start != dirty_stop:
            dirty_start = min(dirty_start, start)
            dirty_stop = max(dirty_stop, stop)
        else:
            dirty_start = start
            dirty_stop = stop
        self.__offscreen_dirty = (dirty_start, dirty_stop)

        # Just like in __draw_offscreen. FIXME: Need this in one place!
        start -= 8
        stop += 8
        self.queue_draw_area(start, 0, stop - start, alloc.height)

    def __draw_from_offscreen(self, ctx):

        if not self.props.visible:
            return

        alloc = self.get_allocation()
        offscreen_width, offscreen_height = self.__offscreen_size
        rect = Gdk.Rectangle()  # TODO: damage region
        rect.x, rect.y, rect.width, rect.height = 0, 0, alloc.width, alloc.height

        # Fill the background (where the offscreen pixmap doesn't fit) with
        # white. This happens after enlarging the window, until all sentinels
        # have finished running.
        if offscreen_width < alloc.width or offscreen_height < alloc.height:
            ctx.rectangle(rect.x, rect.y, rect.width, rect.height)
            ctx.clip()

            if offscreen_width < alloc.width:
                ctx.rectangle(
                    offscreen_width, 0, alloc.width, offscreen_height)
            if offscreen_height < alloc.height:
                ctx.new_path()
                ctx.rectangle(0, offscreen_height, alloc.width, alloc.height)

            ctx.set_line_width(0.)
            ctx.set_source_rgb(1., 1., 1.)
            ctx.fill()

        ctx.set_source_surface(self.__offscreen)
        ctx.rectangle(rect.x, rect.y, rect.width, rect.height)
        ctx.paint()

        self.__draw_position(ctx, clip=rect)

    def update(self, model):

        self.clear()
        self.model = model

        if model is not None:
            self.__dist_sentinel_progress = 0
            self.process.freq_sentinel = LineFrequencySentinel(model)
            self.process.dist_sentinel = LevelDistributionSentinel(
                self.process.freq_sentinel, model)
            width = self.get_allocation().width
            self.process.freq_sentinel.run_for(width)
            self.process.run()

    def clear(self):

        self.model = None
        self.process.abort()
        self.process.freq_sentinel = None
        self.process.dist_sentinel = None
        self.__invalidate_offscreen(0, -1)

    def update_position(self, start_ts, end_ts):

        if not self.process.freq_sentinel:
            return

        if not self.process.freq_sentinel.data:
            return

        alloc = self.get_allocation()

        # Queue old position rectangle for redraw:
        if self.__position_ts_range is not None:
            start, stop = self.ts_range_to_position(*self.__position_ts_range)
            self.queue_draw_area(start - 1, 0, stop - start + 2, alloc.height)
        # And the new one:
        start, stop = self.ts_range_to_position(start_ts, end_ts)
        self.queue_draw_area(start - 1, 0, stop - start + 2, alloc.height)

        self.__position_ts_range = (start_ts, end_ts,)

    def find_indicative_time_step(self):

        MINIMUM_PIXEL_STEP = 32
        time_per_pixel = self.process.freq_sentinel.step
        return 32  # FIXME use self.freq_sentinel.step and len (self.process.freq_sentinel.data)

    def __draw_offscreen(self):

        dirty_start, dirty_stop = self.__offscreen_dirty
        if dirty_start == dirty_stop:
            return

        self.__offscreen_dirty = (0, 0)
        width, height = self.__offscreen_size

        ctx = cairo.Context(self.__offscreen)

        # Indicator (triangle) size is 8, so we need to draw surrounding areas
        # a bit:
        dirty_start -= 8
        dirty_stop += 8
        dirty_start = max(dirty_start, 0)
        dirty_stop = min(dirty_stop, width)

        ctx.rectangle(dirty_start, 0., dirty_stop, height)
        ctx.clip()

        # White background rectangle.
        ctx.set_line_width(0.)
        ctx.rectangle(0, 0, width, height)
        ctx.set_source_rgb(1., 1., 1.)
        ctx.fill()
        ctx.new_path()

        # Horizontal reference lines.
        ctx.set_line_width(1.)
        ctx.set_source_rgb(.95, .95, .95)
        for i in range(height // 16):
            y = i * 16 - .5
            ctx.move_to(0, y)
            ctx.line_to(width, y)
            ctx.stroke()

        if self.process.freq_sentinel is None:
            return

        # Vertical reference lines.
        pixel_step = self.find_indicative_time_step()
        ctx.set_source_rgb(.9, .9, .9)
        start = dirty_start - dirty_start % pixel_step
        for x in range(start + pixel_step, dirty_stop, pixel_step):
            ctx.move_to(x - .5, 0)
            ctx.line_to(x - .5, height)
            ctx.stroke()

        if not self.process.freq_sentinel.data:
            self.logger.debug("frequency sentinel has no data yet")
            return

        ctx.translate(dirty_start, 0.)

        maximum = max(self.process.freq_sentinel.data)

        ctx.set_source_rgb(0., 0., 0.)
        data = self.process.freq_sentinel.data[dirty_start:dirty_stop]
        self.__draw_graph(ctx, height, maximum, data)

        if not self.process.dist_sentinel.data:
            self.logger.debug("level distribution sentinel has no data yet")
            return

        colors = LevelColorThemeTango().colors
        dist_data = self.process.dist_sentinel.data[dirty_start:dirty_stop]

        def cumulative_level_counts(*levels):
            for level_counts in dist_data:
                yield sum((level_counts[level] for level in levels))

        level = Data.debug_level_info
        levels_prev = (Data.debug_level_trace,
                       Data.debug_level_fixme,
                       Data.debug_level_log,
                       Data.debug_level_debug,)
        ctx.set_source_rgb(*(colors[level][1].float_tuple()))
        self.__draw_graph(ctx, height, maximum,
                          list(cumulative_level_counts(level, *levels_prev)))

        level = Data.debug_level_debug
        levels_prev = (Data.debug_level_trace,
                       Data.debug_level_fixme,
                       Data.debug_level_log,)
        ctx.set_source_rgb(*(colors[level][1].float_tuple()))
        self.__draw_graph(ctx, height, maximum,
                          list(cumulative_level_counts(level, *levels_prev)))

        level = Data.debug_level_log
        levels_prev = (Data.debug_level_trace, Data.debug_level_fixme,)
        ctx.set_source_rgb(*(colors[level][1].float_tuple()))
        self.__draw_graph(ctx, height, maximum,
                          list(cumulative_level_counts(level, *levels_prev)))

        level = Data.debug_level_fixme
        levels_prev = (Data.debug_level_trace,)
        ctx.set_source_rgb(*(colors[level][1].float_tuple()))
        self.__draw_graph(ctx, height, maximum,
                          list(cumulative_level_counts(level, *levels_prev)))

        level = Data.debug_level_trace
        ctx.set_source_rgb(*(colors[level][1].float_tuple()))
        self.__draw_graph(ctx, height, maximum, [
                          counts[level] for counts in dist_data])

        # Draw error and warning triangle indicators:

        def triangle(ctx, size=8):
            ctx.move_to(-size // 2, 0)
            ctx.line_to((size + 1) // 2, 0)
            ctx.line_to(0, size / 1.41)
            ctx.close_path()

        for level in (Data.debug_level_warning, Data.debug_level_error,):
            ctx.set_source_rgb(*(colors[level][1].float_tuple()))
            for i, counts in enumerate(dist_data):
                if counts[level] == 0:
                    continue
                ctx.translate(i, 0.)
                triangle(ctx)
                ctx.fill()
                ctx.translate(-i, 0.)

    def __draw_graph(self, ctx, height, maximum, data):

        if not data:
            return

        if maximum:
            heights = [height * float(d) / maximum for d in data]
        else:
            heights = [0. for d in data]

        ctx.move_to(0, height)
        for i in range(len(heights)):
            ctx.line_to(i - .5, height - heights[i] + .5)

        ctx.line_to(i, height)
        ctx.close_path()

        ctx.fill()

    def __have_position(self):

        if ((self.process is not None) and
            (self.process.freq_sentinel is not None) and
                (self.process.freq_sentinel.ts_range is not None)):
            return True
        else:
            return False

    def ts_range_to_position(self, start_ts, end_ts):

        if not self.__have_position():
            return (0, 0)

        first_ts, last_ts = self.process.freq_sentinel.ts_range
        step = self.process.freq_sentinel.step
        if step == 0:
            return (0, 0)

        position1 = int(float(start_ts - first_ts) / step)
        position2 = int(float(end_ts - first_ts) / step)

        return (position1, position2)

    def __draw_position(self, ctx, clip=None):

        if not self.__have_position() or self.__position_ts_range is None:
            if not self.__have_position():
                self.logger.debug("have no positions")
            else:
                self.logger.debug("have no positions_ts_range")
            return

        start_ts, end_ts = self.__position_ts_range
        position1, position2 = self.ts_range_to_position(start_ts, end_ts)

        if clip:
            if clip.x + clip.width < position1 - 1 or clip.x > position2 + 1:
                self.logger.debug(
                    "outside of clip range: %d + %d, pos: %d, %d", clip.x, clip.width, position1, position2)
                return
            ctx.rectangle(clip.x, clip.y, clip.width, clip.height)
            ctx.clip()

        height = self.get_allocation().height

        line_width = position2 - position1
        if line_width <= 1:
            ctx.set_source_rgb(1., 0., 0.)
            ctx.set_line_width(1.)
            ctx.move_to(position1 + .5, 0)
            ctx.line_to(position1 + .5, height)
            ctx.stroke()
        else:
            ctx.set_source_rgba(1., 0., 0., .5)
            ctx.rectangle(position1, 0, line_width, height)
            ctx.fill()

    def do_draw(self, cr):

        self.__ensure_offscreen()
        self.__draw_offscreen()
        self.__draw_from_offscreen(cr)

        return True

    def do_configure_event(self, event):

        self.logger.debug("widget size configured to %ix%i",
                          event.width, event.height)

        if event.width < 16:
            return False

        self.update(self.model)

        return False

    def do_get_preferred_height(self):

        return 64, 64  # FIXME:

    def do_button_press_event(self, event):

        if event.button != 1:
            return False

        # TODO: Check if clicked inside a warning/error indicator triangle and
        # navigate there.

        if not self.has_grab():
            self.grab_add()
            self.props.has_tooltip = False

        pos = int(event.x)
        self.emit("change-position", pos)
        return True

    def do_button_release_event(self, event):

        if event.button != 1:
            return False

        if self.has_grab():
            self.grab_remove()
            self.props.has_tooltip = True

        return True

    def do_motion_notify_event(self, event):

        if event.get_state() & Gdk.ModifierType.BUTTON1_MASK:
            self.emit("change-position", int(event.x))
            Gdk.event_request_motions(event)
            return True
        else:
            self._handle_motion(event.x, event.y)
            Gdk.event_request_motions(event)
            return False

    def _handle_motion(self, x, y):

        # TODO: Prelight warning and error indicator triangles.

        pass


class AttachedWindow (object):

    def __init__(self, feature, window):

        self.window = window

        ui = window.ui_manager

        ui.insert_action_group(feature.action_group, 0)

        self.merge_id = ui.new_merge_id()
        ui.add_ui(self.merge_id, "/menubar/ViewMenu/ViewMenuAdditions",
                  "ViewTimeline", "show-timeline",
                  Gtk.UIManagerItemType.MENUITEM, False)

        ui.add_ui(self.merge_id, "/", "TimelineContextMenu", None,
                  Gtk.UIManagerItemType.POPUP, False)
        # TODO: Make hide before/after operate on the partition that the mouse
        # is pointed at instead of the currently selected line.
        # ui.add_ui (self.merge_id, "/TimelineContextMenu", "TimelineHideLinesBefore",
        #            "hide-before-line", Gtk.UIManagerItemType.MENUITEM, False)
        # ui.add_ui (self.merge_id, "/TimelineContextMenu", "TimelineHideLinesAfter",
        #            "hide-after-line", Gtk.UIManagerItemType.MENUITEM, False)
        ui.add_ui(
            self.merge_id, "/TimelineContextMenu", "TimelineShowHiddenLines",
            "show-hidden-lines", Gtk.UIManagerItemType.MENUITEM, False)

        box = window.get_top_attach_point()

        self.timeline = TimelineWidget()
        self.timeline.connect("change-position",
                              self.handle_timeline_change_position)
        box.pack_start(self.timeline, False, False, 0)
        self.timeline.hide()

        self.popup = ui.get_widget("/TimelineContextMenu")
        Common.GUI.widget_add_popup_menu(self.timeline, self.popup)

        box = window.get_side_attach_point()

        self.vtimeline = VerticalTimelineWidget(self.window.log_view)
        box.pack_start(self.vtimeline, False, False, 0)
        self.vtimeline.hide()

        handler = self.handle_log_view_adjustment_value_changed
        adjustment = window.widgets.log_view_scrolled_window.props.vadjustment
        adjustment.connect("value-changed", handler)

        handler = self.handle_show_action_toggled
        action = feature.action_group.get_action("show-timeline")
        action.connect("toggled", handler)
        handler(action)

        handler = self.handle_log_view_notify_model
        self.notify_model_id = window.log_view.connect(
            "notify::model", handler)

        self.idle_scroll_path = None
        self.idle_scroll_id = None

    def detach(self, feature):

        self.window.log_view.disconnect(self.notify_model_id)
        self.notify_model_id = None

        self.window.ui_manager.remove_ui(self.merge_id)
        self.merge_id = None

        self.window.ui_manager.remove_action_group(feature.action_group)

        self.timeline.destroy()
        self.timeline = None

        self.idle_scroll_path = None
        if self.idle_scroll_id is not None:
            GObject.source_remove(self.idle_scroll_id)
            self.idle_scroll_id = None

    def handle_detach_log_file(self, log_file):

        self.timeline.clear()
        self.vtimeline.clear()

    def handle_log_view_notify_model(self, view, gparam):

        model = view.get_model()

        if model is None:
            self.timeline.clear()
            self.vtimeline.clear()
            return

        self.timeline.update(model)

        # Need to dispatch these idly with a low priority to avoid triggering a
        # warning in treeview.get_visible_range:
        def idle_update():
            self.update_timeline_position()
            self.vtimeline.update()
            return False
        GObject.idle_add(idle_update, priority=GObject.PRIORITY_LOW)

    def handle_log_view_adjustment_value_changed(self, adj):

        # FIXME: If not visible, disconnect this handler!
        if not self.timeline.props.visible:
            return

        self.update_timeline_position()
        self.vtimeline.update()

    def update_timeline_position(self):

        visible_range = self.window.get_range()
        if visible_range is None:
            return
        ts1, ts2 = visible_range
        self.timeline.update_position(ts1, ts2)

    def handle_show_action_toggled(self, action):

        show = action.props.active

        if show:
            self.timeline.show()
            self.vtimeline.show()
        else:
            self.timeline.hide()
            self.vtimeline.hide()

    def handle_timeline_change_position(self, widget, pos):

        self.goto_time_position(pos)

    def goto_time_position(self, pos):

        if not self.timeline.process.freq_sentinel:
            return True

        data = self.timeline.process.freq_sentinel.data
        if not data:
            return True

        if pos < 0:
            pos = 0
        elif pos >= len(data):
            pos = len(data) - 1

        count = sum(data[:pos + 1])

        path = (count,)
        self.idle_scroll_path = path

        if self.idle_scroll_id is None:
            self.idle_scroll_id = GObject.idle_add(self.idle_scroll)

        return False

    def idle_scroll(self):

        self.idle_scroll_id = None

        if self.idle_scroll_path is None:
            return False

        path = self.idle_scroll_path
        self.idle_scroll_path = None

        view = self.window.log_view
        view.scroll_to_cell(path, use_align=True, row_align=.5)

        return False


class TimelineFeature (FeatureBase):

    def __init__(self, app):

        self.logger = logging.getLogger("ui.timeline")

        self.action_group = Gtk.ActionGroup("TimelineActions")
        self.action_group.add_toggle_actions([("show-timeline",
                                               None, _("_Timeline"),
                                               "<Ctrl>t")])

        self.state = app.state.sections[TimelineState._name]

        self.attached_windows = {}

        action = self.action_group.get_action("show-timeline")
        action.props.active = self.state.shown
        action.connect("toggled", self.handle_show_action_toggled)

    def handle_show_action_toggled(self, action):

        self.state.shown = action.props.active

    def handle_attach_window(self, window):

        self.attached_windows[window] = AttachedWindow(self, window)

    def handle_detach_window(self, window):

        attached_window = self.attached_windows.pop(window)
        attached_window.detach(self)

    def handle_attach_log_file(self, window, log_file):

        pass

    def handle_detach_log_file(self, window, log_file):

        attached_window = self.attached_windows[window]
        attached_window.handle_detach_log_file(log_file)


class TimelineState (Common.GUI.StateSection):

    _name = "timeline"

    shown = Common.GUI.StateBool("shown", default=True)


class Plugin (PluginBase):

    features = (TimelineFeature,)

    def __init__(self, app):

        app.state.add_section_class(TimelineState)
        self.state = app.state.sections[TimelineState._name]



# -*- coding: utf-8; mode: python; -*-
#
#  GStreamer Debug Viewer - View and analyze GStreamer debug log files
#
#  Copyright (C) 2007 René Stadler <mail@renestadler.de>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 3 of the License, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  this program.  If not, see <http://www.gnu.org/licenses/>.

"""GStreamer Debug Viewer GUI module."""

from array import array
from bisect import bisect_left
import logging

from gi.repository import GObject
from gi.repository import Gtk

from GstDebugViewer import Common, Data


class LogModelBase (Common.GUI.GenericTreeModel, metaclass=Common.GUI.MetaModel):

    columns = ("COL_TIME", GObject.TYPE_UINT64,
               "COL_PID", int,
               "COL_THREAD", GObject.TYPE_UINT64,
               "COL_LEVEL", object,
               "COL_CATEGORY", str,
               "COL_FILENAME", str,
               "COL_LINE_NUMBER", int,
               "COL_FUNCTION", str,
               "COL_OBJECT", str,
               "COL_MESSAGE", str,)

    def __init__(self):

        Common.GUI.GenericTreeModel.__init__(self)

        # self.props.leak_references = False

        self.line_offsets = array("I")
        self.line_levels = []  # FIXME: Not so nice!
        self.line_cache = {}

    def ensure_cached(self, line_offset):

        raise NotImplementedError("derived classes must override this method")

    def access_offset(self, offset):

        raise NotImplementedError("derived classes must override this method")

    def iter_rows_offset(self):

        ensure_cached = self.ensure_cached
        line_cache = self.line_cache
        line_levels = self.line_levels
        COL_LEVEL = self.COL_LEVEL
        COL_MESSAGE = self.COL_MESSAGE
        access_offset = self.access_offset

        for i, offset in enumerate(self.line_offsets):
            ensure_cached(offset)
            row = line_cache[offset]
            # adjust special rows
            row[COL_LEVEL] = line_levels[i]
            msg_offset = row[COL_MESSAGE]
            row[COL_MESSAGE] = access_offset(offset + msg_offset)
            yield (row, offset,)
            row[COL_MESSAGE] = msg_offset

    def on_get_flags(self):

        flags = Gtk.TreeModelFlags.LIST_ONLY | Gtk.TreeModelFlags.ITERS_PERSIST

        return flags

    def on_get_n_columns(self):

        return len(self.column_types)

    def on_get_column_type(self, col_id):

        return self.column_types[col_id]

    def on_get_iter(self, path):

        if not path:
            return

        if len(path) > 1:
            # Flat model.
            return None

        line_index = path[0]

        if line_index > len(self.line_offsets) - 1:
            return None

        return line_index

    def on_get_path(self, rowref):

        line_index = rowref

        return (line_index,)

    def on_get_value(self, line_index, col_id):

        last_index = len(self.line_offsets) - 1

        if line_index > last_index:
            return None

        if col_id == self.COL_LEVEL:
            return self.line_levels[line_index]

        line_offset = self.line_offsets[line_index]
        self.ensure_cached(line_offset)

        value = self.line_cache[line_offset][col_id]
        if col_id == self.COL_MESSAGE:
            # strip whitespace + newline
            value = self.access_offset(line_offset + value).strip()
        elif col_id in (self.COL_TIME, self.COL_THREAD):
            value = GObject.Value(GObject.TYPE_UINT64, value)

        return value

    def get_value_range(self, col_id, start, stop):

        if col_id != self.COL_LEVEL:
            raise NotImplementedError("XXX FIXME")

        return self.line_levels[start:stop]

    def on_iter_next(self, line_index):

        last_index = len(self.line_offsets) - 1

        if line_index >= last_index:
            return None
        else:
            return line_index + 1

    def on_iter_children(self, parent):

        return self.on_iter_nth_child(parent, 0)

    def on_iter_has_child(self, rowref):

        return False

    def on_iter_n_children(self, rowref):

        if rowref is not None:
            return 0

        return len(self.line_offsets)

    def on_iter_nth_child(self, parent, n):

        last_index = len(self.line_offsets) - 1

        if parent or n > last_index:
            return None

        return n

    def on_iter_parent(self, child):

        return None

    # def on_ref_node (self, rowref):

    # pass

    # def on_unref_node (self, rowref):

    # pass


class LazyLogModel (LogModelBase):

    def __init__(self, log_obj=None):

        LogModelBase.__init__(self)

        self.__log_obj = log_obj

        if log_obj:
            self.set_log(log_obj)

    def set_log(self, log_obj):

        self.__fileobj = log_obj.fileobj

        self.line_cache.clear()
        self.line_offsets = log_obj.line_cache.offsets
        self.line_levels = log_obj.line_cache.levels

    def access_offset(self, offset):

        # TODO: Implement using one slice access instead of seek+readline.
        self.__fileobj.seek(offset)
        return self.__fileobj.readline()

    def ensure_cached(self, line_offset):

        if line_offset in self.line_cache:
            return

        if len(self.line_cache) > 10000:
            self.line_cache.clear()

        self.__fileobj.seek(line_offset)
        line = self.__fileobj.readline()

        self.line_cache[line_offset] = Data.LogLine.parse_full(line)


class FilteredLogModelBase (LogModelBase):

    def __init__(self, super_model):

        LogModelBase.__init__(self)

        self.logger = logging.getLogger("filter-model-base")

        self.super_model = super_model
        self.access_offset = super_model.access_offset
        self.ensure_cached = super_model.ensure_cached
        self.line_cache = super_model.line_cache

    def line_index_to_super(self, line_index):

        raise NotImplementedError("index conversion not supported")

    def line_index_from_super(self, super_line_index):

        raise NotImplementedError("index conversion not supported")


class FilteredLogModel (FilteredLogModelBase):

    def __init__(self, super_model):

        FilteredLogModelBase.__init__(self, super_model)

        self.logger = logging.getLogger("filtered-log-model")

        self.filters = []
        self.reset()
        self.__active_process = None
        self.__filter_progress = 0.

    def reset(self):

        self.logger.debug("reset filter")

        self.line_offsets = self.super_model.line_offsets
        self.line_levels = self.super_model.line_levels
        self.super_index = range(len(self.line_offsets))

        del self.filters[:]

    def __filter_process(self, filter):

        YIELD_LIMIT = 10000

        self.logger.debug("preparing new filter")
        new_line_offsets = array("I")
        new_line_levels = []
        new_super_index = array("I")
        level_id = self.COL_LEVEL
        func = filter.filter_func

        def enum():
            i = 0
            for row, offset in self.iter_rows_offset():
                line_index = self.super_index[i]
                yield (line_index, row, offset,)
                i += 1
        self.logger.debug("running filter")
        progress = 0.
        progress_full = float(len(self))
        y = YIELD_LIMIT
        for i, row, offset in enum():
            if func(row):
                new_line_offsets.append(offset)
                new_line_levels.append(row[level_id])
                new_super_index.append(i)
            y -= 1
            if y == 0:
                progress += float(YIELD_LIMIT)
                self.__filter_progress = progress / progress_full
                y = YIELD_LIMIT
                yield True
        self.line_offsets = new_line_offsets
        self.line_levels = new_line_levels
        self.super_index = new_super_index
        self.logger.debug("filtering finished")

        self.__filter_progress = 1.
        self.__handle_filter_process_finished()
        yield False

    def add_filter(self, filter, dispatcher):

        if self.__active_process is not None:
            raise ValueError("dispatched a filter process already")

        self.logger.debug("adding filter")

        self.filters.append(filter)

        self.__dispatcher = dispatcher
        self.__active_process = self.__filter_process(filter)
        dispatcher(self.__active_process)

    def abort_process(self):

        if self.__active_process is None:
            raise ValueError("no filter process running")

        self.__dispatcher.cancel()
        self.__active_process = None
        self.__dispatcher = None

        del self.filters[-1]

    def get_filter_progress(self):

        if self.__active_process is None:
            raise ValueError("no filter process running")

        return self.__filter_progress

    def __handle_filter_process_finished(self):

        self.__active_process = None
        self.handle_process_finished()

    def handle_process_finished(self):

        pass

    def line_index_from_super(self, super_line_index):

        return bisect_left(self.super_index, super_line_index)

    def line_index_to_super(self, line_index):

        return self.super_index[line_index]

    def set_range(self, super_start, super_stop):

        old_super_start = self.line_index_to_super(0)
        old_super_stop = self.line_index_to_super(
            len(self.super_index) - 1) + 1

        self.logger.debug("set range (%i, %i), current (%i, %i)",
                          super_start, super_stop, old_super_start, old_super_stop)

        if len(self.filters) == 0:
            # Identity.
            self.super_index = range(super_start, super_stop)
            self.line_offsets = SubRange(self.super_model.line_offsets,
                                         super_start, super_stop)
            self.line_levels = SubRange(self.super_model.line_levels,
                                        super_start, super_stop)
            return

        if super_start < old_super_start:
            # TODO:
            raise NotImplementedError("Only handling further restriction of the range"
                                      " (start offset = %i)" % (super_start,))

        if super_stop > old_super_stop:
            # TODO:
            raise NotImplementedError("Only handling further restriction of the range"
                                      " (end offset = %i)" % (super_stop,))

        start = self.line_index_from_super(super_start)
        stop = self.line_index_from_super(super_stop)

        self.super_index = SubRange(self.super_index, start, stop)
        self.line_offsets = SubRange(self.line_offsets, start, stop)
        self.line_levels = SubRange(self.line_levels, start, stop)


class SubRange (object):

    __slots__ = ("size", "start", "stop",)

    def __init__(self, size, start, stop):

        if start > stop:
            raise ValueError(
                "need start <= stop (got %r, %r)" % (start, stop,))

        if isinstance(size, type(self)):
            # Another SubRange, don't stack:
            start += size.start
            stop += size.start
            size = size.size

        self.size = size
        self.start = start
        self.stop = stop

    def __getitem__(self, i):

        if isinstance(i, slice):
            stop = i.stop
            if stop >= 0:
                stop += self.start
            else:
                stop += self.stop

            return self.size[i.start + self.start:stop]
        else:
            return self.size[i + self.start]

    def __len__(self):

        return self.stop - self.start

    def __iter__(self):

        size = self.size
        for i in range(self.start, self.stop):
            yield size[i]


class LineViewLogModel (FilteredLogModelBase):

    def __init__(self, super_model):

        FilteredLogModelBase.__init__(self, super_model)

        self.line_offsets = []
        self.line_levels = []

        self.parent_indices = []

    def reset(self):

        del self.line_offsets[:]
        del self.line_levels[:]

    def line_index_to_super(self, line_index):

        return self.parent_indices[line_index]

    def insert_line(self, position, super_line_index):

        if position == -1:
            position = len(self.line_offsets)
        li = super_line_index
        self.line_offsets.insert(position, self.super_model.line_offsets[li])
        self.line_levels.insert(position, self.super_model.line_levels[li])
        self.parent_indices.insert(position, super_line_index)

        path = (position,)
        tree_iter = self.get_iter(path)
        self.row_inserted(path, tree_iter)

    def replace_line(self, line_index, super_line_index):

        li = line_index
        self.line_offsets[li] = self.super_model.line_offsets[super_line_index]
        self.line_levels[li] = self.super_model.line_levels[super_line_index]
        self.parent_indices[li] = super_line_index

        path = (line_index,)
        tree_iter = self.get_iter(path)
        self.row_changed(path, tree_iter)

    def remove_line(self, line_index):

        for l in (self.line_offsets,
                  self.line_levels,
                  self.parent_indices,):
            del l[line_index]

        path = (line_index,)
        self.row_deleted(path)



# -*- coding: utf-8; mode: python; -*-
#
#  GStreamer Development Utilities
#
#  Copyright (C) 2007 René Stadler <mail@renestadler.de>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 3 of the License, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  this program; if not, see <http://www.gnu.org/licenses/>.

"""GStreamer Development Utilities Common utils module."""

import os
import logging
import subprocess as _subprocess


class SingletonMeta (type):

    def __init__(cls, name, bases, dict_):

        from weakref import WeakValueDictionary

        super(SingletonMeta, cls).__init__(name, bases, dict_)

        cls._singleton_instances = WeakValueDictionary()

    def __call__(cls, *a, **kw):

        kw_key = tuple(sorted(kw.items()))

        try:
            obj = cls._singleton_instances[a + kw_key]
        except KeyError:
            obj = super(SingletonMeta, cls).__call__(*a, **kw)
            cls._singleton_instances[a + kw_key] = obj
        return obj


def gettext_cache():
    """Return a callable object that operates like gettext.gettext, but is much
    faster when a string is looked up more than once.  This is very useful in
    loops, where calling gettext.gettext can quickly become a major performance
    bottleneck."""

    from gettext import gettext

    d = {}

    def gettext_cache_access(s):

        if s not in d:
            d[s] = gettext(s)
        return d[s]

    return gettext_cache_access


class ClassProperty (property):

    "Like the property class, but also invokes the getter for class access."

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):

        property.__init__(self, fget, fset, fdel, doc)

        self.__fget = fget

    def __get__(self, obj, obj_class=None):

        ret = property.__get__(self, obj, obj_class)
        if ret == self:
            return self.__fget(None)
        else:
            return ret


class _XDGClass (object):

    """Partial implementation of the XDG Base Directory specification v0.6.

    http://standards.freedesktop.org/basedir-spec/basedir-spec-0.6.html"""

    def __init__(self):

        self._add_base_dir("DATA_HOME", "~/.local/share")
        self._add_base_dir("CONFIG_HOME", "~/.config")
        self._add_base_dir("CACHE_HOME", "~/.cache")

    def _add_base_dir(self, name, default):

        dir = os.environ.get("XDG_%s" % (name,))
        if not dir:
            dir = os.path.expanduser(os.path.join(*default.split("/")))

        setattr(self, name, dir)


XDG = _XDGClass()


class SaveWriteFile (object):

    def __init__(self, filename, mode="wt"):

        from tempfile import mkstemp

        self.logger = logging.getLogger("tempfile")

        dir = os.path.dirname(filename)
        base_name = os.path.basename(filename)
        temp_prefix = "%s-tmp" % (base_name,)

        if dir:
            # Destination dir differs from current directory, ensure that it
            # exists:
            try:
                os.makedirs(dir)
            except OSError:
                pass

            self.clean_stale(dir, temp_prefix)

        fd, temp_name = mkstemp(dir=dir, prefix=temp_prefix)

        self.target_name = filename
        self.temp_name = temp_name
        self.real_file = os.fdopen(fd, mode)

    def __enter__(self):

        return self

    def __exit__(self, *exc_args):

        if exc_args == (None, None, None,):
            self.close()
        else:
            self.discard()

    def __del__(self):

        try:
            self.discard()
        except AttributeError:
            # If __init__ failed, self has no real_file attribute.
            pass

    def __close_real(self):

        if self.real_file:
            self.real_file.close()
            self.real_file = None

    def clean_stale(self, dir, temp_prefix):

        from time import time
        from glob import glob

        now = time()
        pattern = os.path.join(dir, "%s*" % (temp_prefix,))

        for temp_filename in glob(pattern):
            mtime = os.stat(temp_filename).st_mtime
            if now - mtime > 3600:
                self.logger.info("deleting stale temporary file %s",
                                 temp_filename)
                try:
                    os.unlink(temp_filename)
                except EnvironmentError as exc:
                    self.logger.warning("deleting stale temporary file "
                                        "failed: %s", exc)

    def tell(self, *a, **kw):

        return self.real_file.tell(*a, **kw)

    def write(self, *a, **kw):

        return self.real_file.write(*a, **kw)

    def close(self):

        self.__close_real()

        if self.temp_name:
            try:
                os.rename(self.temp_name, self.target_name)
            except OSError as exc:
                import errno
                if exc.errno == errno.EEXIST:
                    # We are probably on windows.
                    os.unlink(self.target_name)
                    os.rename(self.temp_name, self.target_name)
            self.temp_name = None

    def discard(self):

        self.__close_real()

        if self.temp_name:

            try:
                os.unlink(self.temp_name)
            except EnvironmentError as exc:
                self.logger.warning("deleting temporary file failed: %s", exc)
            self.temp_name = None


class TeeWriteFile (object):

    # TODO Py2.5: Add context manager methods.

    def __init__(self, *file_objects):

        self.files = list(file_objects)

    def close(self):

        for file in self.files:
            file.close()

    def flush(self):

        for file in self.files:
            file.flush()

    def write(self, string):

        for file in self.files:
            file.write(string)

    def writelines(self, lines):

        for file in self.files:
            file.writelines(lines)


class FixedPopen (_subprocess.Popen):

    def __init__(self, args, **kw):

        # Unconditionally specify all descriptors as redirected, to
        # work around Python bug #1358527 (which is triggered for
        # console-less applications on Windows).

        close = []

        for name in ("stdin", "stdout", "stderr",):
            target = kw.get(name)
            if not target:
                kw[name] = _subprocess.PIPE
                close.append(name)

        _subprocess.Popen.__init__(self, args, **kw)

        for name in close:
            fp = getattr(self, name)
            fp.close()
            setattr(self, name, None)


class DevhelpError (EnvironmentError):

    pass


class DevhelpUnavailableError (DevhelpError):

    pass


class DevhelpClient (object):

    def available(self):

        try:
            self.version()
        except DevhelpUnavailableError:
            return False
        else:
            return True

    def version(self):

        return self._invoke("--version")

    def search(self, entry):

        self._invoke_no_interact("-s", entry)

    def _check_os_error(self, exc):

        import errno
        if exc.errno == errno.ENOENT:
            raise DevhelpUnavailableError()

    def _invoke(self, *args):

        from subprocess import PIPE

        try:
            proc = FixedPopen(("devhelp",) + args,
                              stdout=PIPE)
        except OSError as exc:
            self._check_os_error(exc)
            raise

        out, err = proc.communicate()

        if proc.returncode is not None and proc.returncode != 0:
            raise DevhelpError("devhelp exited with status %i"
                               % (proc.returncode,))
        return out

    def _invoke_no_interact(self, *args):

        from subprocess import PIPE

        try:
            proc = FixedPopen(("devhelp",) + args)
        except OSError as exc:
            self._check_os_error(exc)
            raise



# -*- coding: utf-8; mode: python; -*-
#
#  GStreamer Debug Viewer - View and analyze GStreamer debug log files
#
#  Copyright (C) 2007 René Stadler <mail@renestadler.de>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 3 of the License, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  this program.  If not, see <http://www.gnu.org/licenses/>.

"""GStreamer Debug Viewer GUI module."""

import logging

from gi.repository import Gtk, GLib

from GstDebugViewer import Common, Data
from GstDebugViewer.GUI.colors import LevelColorThemeTango
from GstDebugViewer.GUI.models import LazyLogModel, LogModelBase


def _(s):
    return s

# Sync with gst-inspector!


class Column (object):

    """A single list view column, managed by a ColumnManager instance."""

    name = None
    id = None
    label_header = None
    get_modify_func = None
    get_data_func = None
    get_sort_func = None

    def __init__(self):

        view_column = Gtk.TreeViewColumn(self.label_header)
        view_column.props.reorderable = True

        self.view_column = view_column


class SizedColumn (Column):

    default_size = None

    def compute_default_size(self):

        return None

# Sync with gst-inspector?


class TextColumn (SizedColumn):

    font_family = None

    def __init__(self):

        Column.__init__(self)

        column = self.view_column
        cell = Gtk.CellRendererText()
        column.pack_start(cell, True)

        cell.props.yalign = 0.
        cell.props.ypad = 0

        if self.font_family:
            cell.props.family = self.font_family
            cell.props.family_set = True

        if self.get_data_func:
            data_func = self.get_data_func()
            assert data_func
            id_ = self.id
            if id_ is not None:
                def cell_data_func(column, cell, model, tree_iter, user_data):
                    data_func(cell.props, model.get_value(tree_iter, id_))
            else:
                cell_data_func = data_func
            column.set_cell_data_func(cell, cell_data_func)
        elif not self.get_modify_func:
            column.add_attribute(cell, "text", self.id)
        else:
            self.update_modify_func(column, cell)

        column.props.resizable = True

    def update_modify_func(self, column, cell):

        modify_func = self.get_modify_func()
        id_ = self.id

        def cell_data_func(column, cell, model, tree_iter, user_data):
            cell.props.text = modify_func(model.get_value(tree_iter, id_))
        column.set_cell_data_func(cell, cell_data_func)

    def compute_default_size(self):

        values = self.get_values_for_size()
        if not values:
            return SizedColumn.compute_default_size(self)

        cell = self.view_column.get_cells()[0]

        if self.get_modify_func is not None:
            format = self.get_modify_func()
        else:
            def identity(x):
                return x
            format = identity
        max_width = 0
        for value in values:
            cell.props.text = format(value)
            x, y, w, h = self.view_column.cell_get_size()
            max_width = max(max_width, w)

        return max_width

    def get_values_for_size(self):

        return ()


class TimeColumn (TextColumn):

    name = "time"
    label_header = _("Time")
    id = LazyLogModel.COL_TIME
    font_family = "monospace"

    def __init__(self, *a, **kw):

        self.base_time = 0

        TextColumn.__init__(self, *a, **kw)

    def get_modify_func(self):

        if self.base_time:
            time_diff_args = Data.time_diff_args
            base_time = self.base_time

            def format_time(value):
                return time_diff_args(value - base_time)
        else:
            time_args = Data.time_args

            def format_time(value):
                # TODO: This is hard coded to omit hours.
                return time_args(value)[2:]

        return format_time

    def get_values_for_size(self):

        values = [0]

        return values

    def set_base_time(self, base_time):

        self.base_time = base_time

        column = self.view_column
        cell = column.get_cells()[0]
        self.update_modify_func(column, cell)


class LevelColumn (TextColumn):

    name = "level"
    label_header = _("L")
    id = LazyLogModel.COL_LEVEL

    def __init__(self):

        TextColumn.__init__(self)

        cell = self.view_column.get_cells()[0]
        cell.props.xalign = .5

    @staticmethod
    def get_modify_func():

        def format_level(value):
            return value.name[0]

        return format_level

    @staticmethod
    def get_data_func():

        theme = LevelColorThemeTango()
        colors = dict((level, tuple((c.gdk_color()
                                     for c in theme.colors[level])),)
                      for level in Data.debug_levels
                      if level != Data.debug_level_none)

        def level_data_func(cell_props, level):
            cell_props.text = level.name[0]
            if level in colors:
                cell_colors = colors[level]
            else:
                cell_colors = (None, None, None,)
            cell_props.foreground_gdk = cell_colors[0]
            cell_props.background_gdk = cell_colors[1]

        return level_data_func

    def get_values_for_size(self):

        values = [Data.debug_level_log, Data.debug_level_debug,
                  Data.debug_level_info, Data.debug_level_warning,
                  Data.debug_level_error, Data.debug_level_memdump]

        return values


class PidColumn (TextColumn):

    name = "pid"
    label_header = _("PID")
    id = LazyLogModel.COL_PID
    font_family = "monospace"

    @staticmethod
    def get_modify_func():

        return str

    def get_values_for_size(self):

        return ["999999"]


class ThreadColumn (TextColumn):

    name = "thread"
    label_header = _("Thread")
    id = LazyLogModel.COL_THREAD
    font_family = "monospace"

    @staticmethod
    def get_modify_func():

        def format_thread(value):
            return "0x%07x" % (value,)

        return format_thread

    def get_values_for_size(self):

        return [int("ffffff", 16)]


class CategoryColumn (TextColumn):

    name = "category"
    label_header = _("Category")
    id = LazyLogModel.COL_CATEGORY

    def get_values_for_size(self):

        return ["GST_LONG_CATEGORY", "somelongelement"]


class CodeColumn (TextColumn):

    name = "code"
    label_header = _("Code")
    id = None

    @staticmethod
    def get_data_func():

        filename_id = LogModelBase.COL_FILENAME
        line_number_id = LogModelBase.COL_LINE_NUMBER

        def filename_data_func(column, cell, model, tree_iter, user_data):
            args = model.get(tree_iter, filename_id, line_number_id)
            cell.props.text = "%s:%i" % args

        return filename_data_func

    def get_values_for_size(self):

        return ["gstsomefilename.c:1234"]


class FunctionColumn (TextColumn):

    name = "function"
    label_header = _("Function")
    id = LazyLogModel.COL_FUNCTION

    def get_values_for_size(self):

        return ["gst_this_should_be_enough"]


class ObjectColumn (TextColumn):

    name = "object"
    label_header = _("Object")
    id = LazyLogModel.COL_OBJECT

    def get_values_for_size(self):

        return ["longobjectname00"]


class MessageColumn (TextColumn):

    name = "message"
    label_header = _("Message")
    id = None

    def __init__(self, *a, **kw):

        self.highlighters = {}

        TextColumn.__init__(self, *a, **kw)

    def get_data_func(self):

        highlighters = self.highlighters
        id_ = LazyLogModel.COL_MESSAGE

        def message_data_func(column, cell, model, tree_iter, user_data):

            msg = model.get_value(tree_iter, id_).decode("utf8", errors="replace")

            if not highlighters:
                cell.props.text = msg
                return

            if len(highlighters) > 1:
                raise NotImplementedError("FIXME: Support more than one...")

            highlighter = list(highlighters.values())[0]
            row = model[tree_iter]
            ranges = highlighter(row)
            if not ranges:
                cell.props.text = msg
            else:
                tags = []
                prev_end = 0
                end = None
                for start, end in ranges:
                    if prev_end < start:
                        tags.append(
                            GLib.markup_escape_text(msg[prev_end:start]))
                    msg_escape = GLib.markup_escape_text(msg[start:end])
                    tags.append("<span foreground=\'#FFFFFF\'"
                                " background=\'#0000FF\'>%s</span>" % (msg_escape,))
                    prev_end = end
                if end is not None:
                    tags.append(GLib.markup_escape_text(msg[end:]))
                cell.props.markup = "".join(tags)

        return message_data_func

    def get_values_for_size(self):

        values = ["Just some good minimum size"]

        return values


class ColumnManager (Common.GUI.Manager):

    column_classes = ()

    @classmethod
    def iter_item_classes(cls):

        return iter(cls.column_classes)

    def __init__(self):

        self.view = None
        self.actions = None
        self.zoom = 1.0
        self.__columns_changed_id = None
        self.columns = []
        self.column_order = list(self.column_classes)

        self.action_group = Gtk.ActionGroup("ColumnActions")

        def make_entry(col_class):
            return ("show-%s-column" % (col_class.name,),
                    None,
                    col_class.label_header,
                    None,
                    None,
                    None,
                    True,)

        entries = [make_entry(cls) for cls in self.column_classes]
        self.action_group.add_toggle_actions(entries)

    def iter_items(self):

        return iter(self.columns)

    def attach(self):

        for col_class in self.column_classes:
            action = self.get_toggle_action(col_class)
            if action.props.active:
                self._add_column(col_class())
            action.connect("toggled",
                           self.__handle_show_column_action_toggled,
                           col_class.name)

        self.__columns_changed_id = self.view.connect("columns-changed",
                                                      self.__handle_view_columns_changed)

    def detach(self):

        if self.__columns_changed_id is not None:
            self.view.disconnect(self.__columns_changed_id)
            self.__columns_changed_id = None

    def attach_sort(self):

        sort_model = self.view.get_model()

        # Inform the sorted tree model of any custom sorting functions.
        for col_class in self.column_classes:
            if col_class.get_sort_func:
                sort_func = col_class.get_sort_func()
                sort_model.set_sort_func(col_class.id, sort_func)

    def enable_sort(self):

        sort_model = self.view.get_model()

        if sort_model:
            self.logger.debug("activating sort")
            sort_model.set_sort_column_id(*self.default_sort)
            self.default_sort = None
        else:
            self.logger.debug("not activating sort (no model set)")

    def disable_sort(self):

        self.logger.debug("deactivating sort")

        sort_model = self.view.get_model()

        self.default_sort = tree_sortable_get_sort_column_id(sort_model)

        sort_model.set_sort_column_id(TREE_SORTABLE_UNSORTED_COLUMN_ID,
                                      Gtk.SortType.ASCENDING)

    def set_zoom(self, scale):

        for column in self.columns:
            cell = column.view_column.get_cells()[0]
            cell.props.scale = scale
            column.view_column.queue_resize()

        self.zoom = scale

    def set_base_time(self, base_time):

        try:
            time_column = self.find_item(name=TimeColumn.name)
        except KeyError:
            return

        time_column.set_base_time(base_time)
        self.size_column(time_column)

    def get_toggle_action(self, column_class):

        action_name = "show-%s-column" % (column_class.name,)
        return self.action_group.get_action(action_name)

    def get_initial_column_order(self):

        return tuple(self.column_classes)

    def _add_column(self, column):

        name = column.name
        pos = self.__get_column_insert_position(column)

        if self.view.props.fixed_height_mode:
            column.view_column.props.sizing = Gtk.TreeViewColumnSizing.FIXED

        cell = column.view_column.get_cells()[0]
        cell.props.scale = self.zoom

        self.columns.insert(pos, column)
        self.view.insert_column(column.view_column, pos)

    def _remove_column(self, column):

        self.columns.remove(column)
        self.view.remove_column(column.view_column)

    def __get_column_insert_position(self, column):

        col_class = self.find_item_class(name=column.name)
        pos = self.column_order.index(col_class)
        before = self.column_order[:pos]
        shown_names = [col.name for col in self.columns]
        for col_class in before:
            if col_class.name not in shown_names:
                pos -= 1
        return pos

    def __iter_next_hidden(self, column_class):

        pos = self.column_order.index(column_class)
        rest = self.column_order[pos + 1:]
        for next_class in rest:
            try:
                self.find_item(name=next_class.name)
            except KeyError:
                # No instance -- the column is hidden.
                yield next_class
            else:
                break

    def __handle_show_column_action_toggled(self, toggle_action, name):

        if toggle_action.props.active:
            try:
                # This should fail.
                column = self.find_item(name=name)
            except KeyError:
                col_class = self.find_item_class(name=name)
                self._add_column(col_class())
            else:
                # Out of sync for some reason.
                return
        else:
            try:
                column = self.find_item(name=name)
            except KeyError:
                # Out of sync for some reason.
                return
            else:
                self._remove_column(column)

    def __handle_view_columns_changed(self, element_view):

        view_columns = element_view.get_columns()
        new_visible = [self.find_item(view_column=column)
                       for column in view_columns]

        # We only care about reordering here.
        if len(new_visible) != len(self.columns):
            return

        if new_visible != self.columns:

            new_order = []
            for column in new_visible:
                col_class = self.find_item_class(name=column.name)
                new_order.append(col_class)
                new_order.extend(self.__iter_next_hidden(col_class))

            names = (column.name for column in new_visible)
            self.logger.debug("visible columns reordered: %s",
                              ", ".join(names))

            self.columns[:] = new_visible
            self.column_order[:] = new_order


class ViewColumnManager (ColumnManager):

    column_classes = (
        TimeColumn, LevelColumn, PidColumn, ThreadColumn, CategoryColumn,
        CodeColumn, FunctionColumn, ObjectColumn, MessageColumn,)

    default_column_classes = (
        TimeColumn, LevelColumn, CategoryColumn, CodeColumn,
        FunctionColumn, ObjectColumn, MessageColumn,)

    def __init__(self, state):

        ColumnManager.__init__(self)

        self.logger = logging.getLogger("ui.columns")

        self.state = state

    def attach(self, view):

        self.view = view
        view.connect("notify::model", self.__handle_notify_model)

        order = self.state.column_order
        if len(order) == len(self.column_classes):
            self.column_order[:] = order

        visible = self.state.columns_visible
        if not visible:
            visible = self.default_column_classes
        for col_class in self.column_classes:
            action = self.get_toggle_action(col_class)
            action.props.active = (col_class in visible)

        ColumnManager.attach(self)

        self.columns_sized = False

    def detach(self):

        self.state.column_order = self.column_order
        self.state.columns_visible = self.columns

        return ColumnManager.detach(self)

    def set_zoom(self, scale):

        ColumnManager.set_zoom(self, scale)

        if self.view is None:
            return

        # Timestamp and log level columns are pretty much fixed size, so resize
        # them back to default on zoom change:
        names = (TimeColumn.name,
                 LevelColumn.name,
                 PidColumn.name,
                 ThreadColumn.name)
        for column in self.columns:
            if column.name in names:
                self.size_column(column)

    def size_column(self, column):

        if column.default_size is None:
            default_size = column.compute_default_size()
        else:
            default_size = column.default_size
        # FIXME: Abstract away fixed size setting in Column class!
        if default_size is None:
            # Dummy fallback:
            column.view_column.props.fixed_width = 50
            self.logger.warning(
                "%s column does not implement default size", column.name)
        else:
            column.view_column.props.fixed_width = default_size

    def _add_column(self, column):

        result = ColumnManager._add_column(self, column)
        self.size_column(column)
        return result

    def _remove_column(self, column):

        column.default_size = column.view_column.props.fixed_width
        return ColumnManager._remove_column(self, column)

    def __handle_notify_model(self, view, gparam):

        if self.columns_sized:
            # Already sized.
            return
        model = self.view.get_model()
        if model is None:
            return
        self.logger.debug("model changed, sizing columns")
        for column in self.iter_items():
            self.size_column(column)
        self.columns_sized = True


class WrappingMessageColumn (MessageColumn):

    def wrap_to_width(self, width):

        col = self.view_column
        col.props.max_width = width
        col.get_cells()[0].props.wrap_width = width
        col.queue_resize()


class LineViewColumnManager (ColumnManager):

    column_classes = (TimeColumn, WrappingMessageColumn,)

    def __init__(self):

        ColumnManager.__init__(self)

    def attach(self, window):

        self.__size_update = None

        self.view = window.widgets.line_view
        self.view.set_size_request(0, 0)
        self.view.connect_after("size-allocate", self.__handle_size_allocate)
        ColumnManager.attach(self)

    def __update_sizes(self):

        view_width = self.view.get_allocation().width
        if view_width == self.__size_update:
            # Prevent endless recursion.
            return

        self.__size_update = view_width

        col = self.find_item(name="time")
        other_width = col.view_column.props.width

        try:
            col = self.find_item(name="message")
        except KeyError:
            return

        width = view_width - other_width
        col.wrap_to_width(width)

    def __handle_size_allocate(self, self_, allocation):

        self.__update_sizes()



# -*- coding: utf-8; mode: python; -*-
#
#  GStreamer Debug Viewer - View and analyze GStreamer debug log files
#
#  Copyright (C) 2007 René Stadler <mail@renestadler.de>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 3 of the License, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  this program.  If not, see <http://www.gnu.org/licenses/>.

"""GStreamer Debug Viewer GUI module."""

import os.path

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')

from gi.repository import GObject
from gi.repository import Gdk
from gi.repository import Gtk

from GstDebugViewer import Common
from GstDebugViewer.GUI.columns import ViewColumnManager
from GstDebugViewer.GUI.window import Window


class AppStateSection (Common.GUI.StateSection):

    _name = "state"

    geometry = Common.GUI.StateInt4("window-geometry")
    maximized = Common.GUI.StateBool("window-maximized")

    column_order = Common.GUI.StateItemList("column-order", ViewColumnManager)
    columns_visible = Common.GUI.StateItemList(
        "columns-visible", ViewColumnManager)

    zoom_level = Common.GUI.StateInt("zoom-level")


class AppState (Common.GUI.State):

    def __init__(self, *a, **kw):

        Common.GUI.State.__init__(self, *a, **kw)

        self.add_section_class(AppStateSection)


class App (object):

    def __init__(self):

        self.attach()

    def load_plugins(self):

        from GstDebugViewer import Plugins

        plugin_classes = list(
            Plugins.load([os.path.dirname(Plugins.__file__)]))
        self.plugins = []
        for plugin_class in plugin_classes:
            plugin = plugin_class(self)
            self.plugins.append(plugin)

    def iter_plugin_features(self):

        for plugin in self.plugins:
            for feature in plugin.features:
                yield feature

    def attach(self):

        config_home = Common.utils.XDG.CONFIG_HOME

        state_filename = os.path.join(
            config_home, "gst-debug-viewer", "state")

        self.state = AppState(state_filename)
        self.state_section = self.state.sections["state"]

        self.load_plugins()

        self.windows = []

        # Apply custom widget stying
        # TODO: check for dark theme
        css = b"""
        @define-color normal_bg_color #FFFFFF;
        @define-color shade_bg_color shade(@normal_bg_color, 0.95);
        #log_view row:nth-child(even) {
            background-color: @normal_bg_color;
        }
        #log_view row:nth-child(odd) {
            background-color: @shade_bg_color;
        }
        #log_view row:selected {
            background-color: #4488FF;
        }
        #log_view {
          -GtkTreeView-horizontal-separator: 0;
          -GtkTreeView-vertical-separator: 1;
          outline-width: 0;
          outline-offset: 0;
        }
        """

        style_provider = Gtk.CssProvider()
        style_provider.load_from_data(css)

        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(),
            style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

        self.open_window()

    def detach(self):

        # TODO: If we take over deferred saving from the inspector, specify now
        # = True here!
        self.state.save()

    def run(self):

        try:
            Common.Main.MainLoopWrapper(Gtk.main, Gtk.main_quit).run()
        except BaseException:
            raise
        else:
            self.detach()

    def open_window(self):

        self.windows.append(Window(self))

    def close_window(self, window):

        self.windows.remove(window)
        if not self.windows:
            # GtkTreeView takes some time to go down for large files.  Let's block
            # until the window is hidden:
            GObject.idle_add(Gtk.main_quit)
            Gtk.main()

            Gtk.main_quit()



# -*- coding: utf-8; mode: python; -*-
#
#  GStreamer Development Utilities
#
#  Copyright (C) 2007 René Stadler <mail@renestadler.de>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 3 of the License, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  this program.  If not, see <http://www.gnu.org/licenses/>.

"""GStreamer Development Utilities Common Main module."""

import sys
import os
import traceback
from operator import attrgetter
import logging
import locale
import gettext
from gettext import gettext as _, ngettext

import gi

from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk


class ExceptionHandler (object):

    exc_types = (Exception,)
    priority = 50
    inherit_fork = True

    _handling_exception = False

    def __call__(self, exc_type, exc_value, exc_traceback):

        raise NotImplementedError(
            "derived classes need to override this method")


class DefaultExceptionHandler (ExceptionHandler):
    exc_types = (BaseException,)
    priority = 0
    inherit_fork = True

    def __init__(self, excepthook):

        ExceptionHandler.__init__(self)

        self.excepthook = excepthook

    def __call__(self, *exc_info):

        return self.excepthook(*exc_info)


class ExitOnInterruptExceptionHandler (ExceptionHandler):

    exc_types = (KeyboardInterrupt,)
    priority = 100
    inherit_fork = False

    exit_status = 2

    def __call__(self, *args):

        print("Interrupt caught, exiting.", file=sys.stderr)

        sys.exit(self.exit_status)


class MainLoopWrapper (ExceptionHandler):

    priority = 95
    inherit_fork = False

    def __init__(self, enter, exit):

        ExceptionHandler.__init__(self)

        self.exc_info = (None,) * 3
        self.enter = enter
        self.exit = exit

    def __call__(self, *exc_info):

        self.exc_info = exc_info
        self.exit()

    def run(self):

        ExceptHookManager.register_handler(self)
        try:
            self.enter()
        finally:
            ExceptHookManager.unregister_handler(self)

        if self.exc_info != (None,) * 3:
            # Re-raise unhandled exception that occured while running the loop.
            exc_type, exc_value, exc_tb = self.exc_info
            raise exc_value


class ExceptHookManagerClass (object):

    def __init__(self):

        self._in_forked_child = False

        self.handlers = []

    def setup(self):

        if sys.excepthook == self.__excepthook:
            raise ValueError("already set up")

        hook = sys.excepthook
        self.__instrument_excepthook()
        self.__instrument_fork()
        self.register_handler(DefaultExceptionHandler(hook))

    def shutdown(self):

        if sys.excepthook != self.__excepthook:
            raise ValueError("not set up")

        self.__restore_excepthook()
        self.__restore_fork()

    def __instrument_excepthook(self):

        hook = sys.excepthook
        self._original_excepthook = hook
        sys.excepthook = self.__excepthook

    def __restore_excepthook(self):

        sys.excepthook = self._original_excepthook

    def __instrument_fork(self):

        try:
            fork = os.fork
        except AttributeError:
            # System has no fork() system call.
            self._original_fork = None
        else:
            self._original_fork = fork
            os.fork = self.__fork

    def __restore_fork(self):

        if not hasattr(os, "fork"):
            return

        os.fork = self._original_fork

    def entered_forked_child(self):

        self._in_forked_child = True

        for handler in tuple(self.handlers):
            if not handler.inherit_fork:
                self.handlers.remove(handler)

    def register_handler(self, handler):

        if self._in_forked_child and not handler.inherit_fork:
            return

        self.handlers.append(handler)

    def unregister_handler(self, handler):

        self.handlers.remove(handler)

    def __fork(self):

        pid = self._original_fork()
        if pid == 0:
            # Child process.
            self.entered_forked_child()
        return pid

    def __excepthook(self, exc_type, exc_value, exc_traceback):

        for handler in sorted(self.handlers,
                              key=attrgetter("priority"),
                              reverse=True):

            if handler._handling_exception:
                continue

            for type_ in handler.exc_types:
                if issubclass(exc_type, type_):
                    break
            else:
                continue

            handler._handling_exception = True
            handler(exc_type, exc_value, exc_traceback)
            # Not using try...finally on purpose here.  If the handler itself
            # fails with an exception, this prevents recursing into it again.
            handler._handling_exception = False
            return

        else:
            from warnings import warn
            warn("ExceptHookManager: unhandled %r" % (exc_value,),
                 RuntimeWarning,
                 stacklevel=2)


ExceptHookManager = ExceptHookManagerClass()


class PathsBase (object):

    data_dir = None
    icon_dir = None
    locale_dir = None

    @classmethod
    def setup_installed(cls, data_prefix):
        """Set up paths for running from a regular installation."""

        pass

    @classmethod
    def setup_devenv(cls, source_dir):
        """Set up paths for running the development environment
        (i.e. directly from the source dist)."""

        pass

    @classmethod
    def ensure_setup(cls):
        """If paths are still not set up, try to set from a fallback."""

        if cls.data_dir is None:
            source_dir = os.path.dirname(
                os.path.dirname(os.path.abspath(__file__)))
            cls.setup_devenv(source_dir)

    def __new__(cls):

        raise RuntimeError("do not create instances of this class -- "
                           "use the class object directly")


class PathsProgramBase (PathsBase):

    program_name = None

    @classmethod
    def setup_installed(cls, data_prefix):

        if cls.program_name is None:
            raise NotImplementedError(
                "derived classes need to set program_name attribute")

        cls.data_dir = os.path.join(data_prefix, cls.program_name)
        cls.icon_dir = os.path.join(data_prefix, "icons")
        cls.locale_dir = os.path.join(data_prefix, "locale")

    @classmethod
    def setup_devenv(cls, source_dir):
        """Set up paths for running the development environment
        (i.e. directly from the source dist)."""

        # This is essential: The GUI module needs to find the .glade file.
        cls.data_dir = os.path.join(source_dir, "data")

        # The locale data might be missing if "setup.py build" wasn't run.
        cls.locale_dir = os.path.join(source_dir, "build", "mo")

        # Not setting icon_dir.  It is not useful since we don't employ the
        # needed directory structure in the source dist.


def _init_excepthooks():

    ExceptHookManager.setup()
    ExceptHookManager.register_handler(ExitOnInterruptExceptionHandler())


def _init_paths(paths):

    paths.ensure_setup()


def _init_locale(gettext_domain=None):

    if Paths.locale_dir and gettext_domain is not None:
        try:
            locale.setlocale(locale.LC_ALL, "")
        except locale.Error as exc:
            from warnings import warn
            warn("locale error: %s" % (exc,),
                 RuntimeWarning,
                 stacklevel=2)
            Paths.locale_dir = None
        else:
            gettext.bindtextdomain(gettext_domain, Paths.locale_dir)
            gettext.textdomain(gettext_domain)


def _init_logging(level):
    if level == "none":
        return

    mapping = {"debug": logging.DEBUG,
               "info": logging.INFO,
               "warning": logging.WARNING,
               "error": logging.ERROR,
               "critical": logging.CRITICAL}
    logging.basicConfig(level=mapping[level],
                        format='%(asctime)s.%(msecs)03d %(levelname)8s %(name)20s: %(message)s',
                        datefmt='%H:%M:%S')

    logger = logging.getLogger("main")
    logger.debug("logging at level %s", logging.getLevelName(level))
    logger.info("using Python %i.%i.%i %s %i", *sys.version_info)


def _init_log_option(parser):
    choices = ["none", "debug", "info", "warning", "error", "critical"]
    parser.add_option("--log-level", "-l",
                      type="choice",
                      choices=choices,
                      action="store",
                      dest="log_level",
                      default="none",
                      help=_("Enable logging, possible values: ") + ", ".join(choices))
    return parser


def main(main_function, option_parser, gettext_domain=None, paths=None):

    # FIXME:
    global Paths
    Paths = paths

    _init_excepthooks()
    _init_paths(paths)
    _init_locale(gettext_domain)
    parser = _init_log_option(option_parser)
    options, args = option_parser.parse_args()
    _init_logging(options.log_level)

    try:
        main_function(args)
    finally:
        logging.shutdown()



# -*- coding: utf-8; mode: python; -*-
#
#  GStreamer Debug Viewer - View and analyze GStreamer debug log files
#
#  Copyright (C) 2007 René Stadler <mail@renestadler.de>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 3 of the License, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  this program.  If not, see <http://www.gnu.org/licenses/>.

"""GStreamer Debug Viewer GUI module."""

from GstDebugViewer.GUI.models import LogModelBase


def get_comparison_function(all_but_this):

    if (all_but_this):
        return lambda x, y: x == y
    else:
        return lambda x, y: x != y


class Filter (object):

    pass


class DebugLevelFilter (Filter):

    only_this, all_but_this, this_and_above = range(3)

    def __init__(self, debug_level, mode=0):

        col_id = LogModelBase.COL_LEVEL
        if mode == self.this_and_above:
            def comparison_function(x, y):
                return x < y
        else:
            comparison_function = get_comparison_function(
                mode == self.all_but_this)

        def filter_func(row):
            return comparison_function(row[col_id], debug_level)
        self.filter_func = filter_func


class CategoryFilter (Filter):

    def __init__(self, category, all_but_this=False):

        col_id = LogModelBase.COL_CATEGORY
        comparison_function = get_comparison_function(all_but_this)

        def category_filter_func(row):
            return comparison_function(row[col_id], category)
        self.filter_func = category_filter_func


class ObjectFilter (Filter):

    def __init__(self, object_, all_but_this=False):

        col_id = LogModelBase.COL_OBJECT
        comparison_function = get_comparison_function(all_but_this)

        def object_filter_func(row):
            return comparison_function(row[col_id], object_)
        self.filter_func = object_filter_func


class FunctionFilter (Filter):

    def __init__(self, function_, all_but_this=False):

        col_id = LogModelBase.COL_FUNCTION
        comparison_function = get_comparison_function(all_but_this)

        def function_filter_func(row):
            return comparison_function(row[col_id], function_)
        self.filter_func = function_filter_func


class ThreadFilter (Filter):

    def __init__(self, thread_, all_but_this=False):

        col_id = LogModelBase.COL_THREAD
        comparison_function = get_comparison_function(all_but_this)

        def thread_filter_func(row):
            return comparison_function(row[col_id], thread_)
        self.filter_func = thread_filter_func


class FilenameFilter (Filter):

    def __init__(self, filename, all_but_this=False):

        col_id = LogModelBase.COL_FILENAME
        comparison_function = get_comparison_function(all_but_this)

        def filename_filter_func(row):
            return comparison_function(row[col_id], filename)
        self.filter_func = filename_filter_func



# -*- coding: utf-8; mode: python; -*-
#
#  GStreamer Debug Viewer - View and analyze GStreamer debug log files
#
#  Copyright (C) 2007 René Stadler <mail@renestadler.de>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 3 of the License, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  this program.  If not, see <http://www.gnu.org/licenses/>.

"""GStreamer Debug Viewer Data module."""

import os
import logging
import re
import sys

# Nanosecond resolution (like Gst.SECOND)
SECOND = 1000000000


def time_args(ts):

    secs = ts // SECOND

    return "%i:%02i:%02i.%09i" % (secs // 60 ** 2,
                                  secs // 60 % 60,
                                  secs % 60,
                                  ts % SECOND,)


def time_diff_args(time_diff):

    if time_diff >= 0:
        sign = "+"
    else:
        sign = "-"

    secs = abs(time_diff) // SECOND

    return "%s%02i:%02i.%09i" % (sign,
                                 secs // 60,
                                 secs % 60,
                                 abs(time_diff) % SECOND,)


def time_args_no_hours(ts):

    secs = ts // SECOND

    return "%02i:%02i.%09i" % (secs // 60,
                               secs % 60,
                               ts % SECOND,)


def parse_time(st):
    """Parse time strings that look like "0:00:00.0000000"."""

    h, m, s = st.split(":")
    secs, subsecs = s.split(".")

    return int((int(h) * 60 ** 2 + int(m) * 60) * SECOND) + \
        int(secs) * SECOND + int(subsecs)


class DebugLevel (int):

    __names = ["NONE", "ERROR", "WARN", "FIXME",
               "INFO", "DEBUG", "LOG", "TRACE", "MEMDUMP"]
    __instances = {}

    def __new__(cls, level):

        try:
            level_int = int(level)
        except (ValueError, TypeError,):
            try:
                level_int = cls.__names.index(level.upper())
            except ValueError:
                raise ValueError("no debug level named %r" % (level,))
        if level_int in cls.__instances:
            return cls.__instances[level_int]
        else:
            new_instance = int.__new__(cls, level_int)
            new_instance.name = cls.__names[level_int]
            cls.__instances[level_int] = new_instance
            return new_instance

    def __repr__(self):

        return "<%s %s (%i)>" % (type(self).__name__, self.__names[self], self,)

    def higher_level(self):

        if self == len(self.__names) - 1:
            raise ValueError("already the highest debug level")

        return DebugLevel(self + 1)

    def lower_level(self):

        if self == 0:
            raise ValueError("already the lowest debug level")

        return DebugLevel(self - 1)


debug_level_none = DebugLevel("NONE")
debug_level_error = DebugLevel("ERROR")
debug_level_warning = DebugLevel("WARN")
debug_level_info = DebugLevel("INFO")
debug_level_debug = DebugLevel("DEBUG")
debug_level_log = DebugLevel("LOG")
debug_level_fixme = DebugLevel("FIXME")
debug_level_trace = DebugLevel("TRACE")
debug_level_memdump = DebugLevel("MEMDUMP")
debug_levels = [debug_level_none,
                debug_level_trace,
                debug_level_fixme,
                debug_level_log,
                debug_level_debug,
                debug_level_info,
                debug_level_warning,
                debug_level_error,
                debug_level_memdump]

# For stripping color codes:
_escape = re.compile(b"\x1b\\[[0-9;]*m")


def strip_escape(s):

    # FIXME: This can be optimized further!

    while b"\x1b" in s:
        s = _escape.sub(b"", s)
    return s


def default_log_line_regex_():

    # "DEBUG             "
    LEVEL = r"([A-Z]+)\s*"
    # "0x8165430 "
    THREAD = r"(0?x?[0-9a-f]+)\s+"  # r"\((0x[0-9a-f]+) - "
    # "0:00:00.777913000  "
    TIME = r"(\d+:\d\d:\d\d\.\d+)\s+"
    CATEGORY = r"([A-Za-z0-9_-]+)\s+"  # "GST_REFCOUNTING ", "flacdec "
    # "  3089 "
    PID = r"(\d+)\s*"
    FILENAME = r"([^:]*):"
    LINE = r"(\d+):"
    FUNCTION = r"(~?[A-Za-z0-9_\s\*,\(\)]*):"
    # FIXME: When non-g(st)object stuff is logged with *_OBJECT (like
    # buffers!), the address is printed *without* <> brackets!
    OBJECT = "(?:<([^>]+)>)?"
    MESSAGE = "(.+)"

    ANSI = "(?:\x1b\\[[0-9;]*m\\s*)*\\s*"

    # New log format:
    expressions = [TIME, ANSI, PID, ANSI, THREAD, ANSI, LEVEL, ANSI,
                   CATEGORY, FILENAME, LINE, FUNCTION, ANSI,
                   OBJECT, ANSI, MESSAGE]
    # Old log format:
    # expressions = [LEVEL, THREAD, TIME, CATEGORY, PID, FILENAME, LINE,
    # FUNCTION, OBJECT, MESSAGE]

    return expressions


def default_log_line_regex():

    return re.compile("".join(default_log_line_regex_()))


class Producer (object):

    def __init__(self):

        self.consumers = []

    def have_load_started(self):

        for consumer in self.consumers:
            consumer.handle_load_started()

    def have_load_finished(self):

        for consumer in self.consumers:
            consumer.handle_load_finished()


class SortHelper (object):

    def __init__(self, fileobj, offsets):

        self._gen = self.__gen(fileobj, offsets)
        next(self._gen)

        # Override in the instance, for performance (this gets called in an
        # inner loop):
        self.find_insert_position = self._gen.send

    @staticmethod
    def find_insert_position(insert_time_string):

        # Stub for documentary purposes.

        pass

    @staticmethod
    def __gen(fileobj, offsets):

        from math import floor
        tell = fileobj.tell
        seek = fileobj.seek
        read = fileobj.read
        time_len = len(time_args(0))

        # We remember the previous insertion point. This gives a nice speed up
        # for larger bubbles which are already sorted. TODO: In practice, log
        # lines only get out of order across threads. Need to check if it pays
        # to parse the thread here and maintain multiple insertion points for
        # heavily interleaved parts of the log.
        pos = 0
        pos_time_string = ""

        insert_pos = None
        while True:
            insert_time_string = (yield insert_pos)

            save_offset = tell()

            if pos_time_string <= insert_time_string:
                lo = pos
                hi = len(offsets)
            else:
                lo = 0
                hi = pos

            # This is a bisection search, except we don't cut the range in the
            # middle each time, but at the 90th percentile. This is because
            # logs are "mostly sorted", so the insertion point is much more
            # likely to be at the end anyways:
            while lo < hi:
                mid = int(floor(lo * 0.1 + hi * 0.9))
                seek(offsets[mid])
                mid_time_string = read(time_len)
                if insert_time_string.encode('utf8') < mid_time_string:
                    hi = mid
                else:
                    lo = mid + 1
            pos = lo
            # Caller will replace row at pos with the new one, so this is
            # correct:
            pos_time_string = insert_time_string

            insert_pos = pos

            seek(save_offset)


class LineCache (Producer):
    """
    offsets: file position for each line
    levels: the debug level for each line
    """

    _lines_per_iteration = 50000

    def __init__(self, fileobj, dispatcher):

        Producer.__init__(self)

        self.logger = logging.getLogger("linecache")
        self.dispatcher = dispatcher

        self.__fileobj = fileobj
        self.__fileobj.seek(0, 2)
        self.__file_size = self.__fileobj.tell()
        self.__fileobj.seek(0)

        self.offsets = []
        self.levels = []  # FIXME

    def start_loading(self):

        self.logger.debug("dispatching load process")
        self.have_load_started()
        self.dispatcher(self.__process())

    def get_progress(self):

        return float(self.__fileobj.tell()) / self.__file_size

    def __process(self):

        offsets = self.offsets
        levels = self.levels

        dict_levels = {"T": debug_level_trace, "F": debug_level_fixme,
                       "L": debug_level_log, "D": debug_level_debug,
                       "I": debug_level_info, "W": debug_level_warning,
                       "E": debug_level_error, " ": debug_level_none,
                       "M": debug_level_memdump, }
        ANSI = "(?:\x1b\\[[0-9;]*m)?"
        ANSI_PATTERN = r"\d:\d\d:\d\d\.\d+ " + ANSI + \
                       r" *\d+" + ANSI + \
                       r" +0?x?[0-9a-f]+ +" + ANSI + \
                       r"([TFLDIEWM ])"
        BARE_PATTERN = ANSI_PATTERN.replace(ANSI, "")
        rexp_bare = re.compile(BARE_PATTERN)
        rexp_ansi = re.compile(ANSI_PATTERN)
        rexp = rexp_bare

        # Moving attribute lookups out of the loop:
        readline = self.__fileobj.readline
        tell = self.__fileobj.tell
        rexp_match = rexp.match
        levels_append = levels.append
        offsets_append = offsets.append
        dict_levels_get = dict_levels.get

        self.__fileobj.seek(0)
        limit = self._lines_per_iteration
        last_line = ""
        i = 0
        sort_helper = SortHelper(self.__fileobj, offsets)
        find_insert_position = sort_helper.find_insert_position
        while True:
            i += 1
            if i >= limit:
                i = 0
                yield True

            offset = tell()
            line = readline().decode('utf-8', errors='replace')
            if not line:
                break
            match = rexp_match(line)
            if match is None:
                if rexp is rexp_ansi or "\x1b" not in line:
                    continue

                match = rexp_ansi.match(line)
                if match is None:
                    continue
                # Switch to slower ANSI parsing:
                rexp = rexp_ansi
                rexp_match = rexp.match

            # Timestamp is in the very beginning of the row, and can be sorted
            # by lexical comparison. That's why we don't bother parsing the
            # time to integer. We also don't have to take a substring here,
            # which would be a useless memcpy.
            if line >= last_line:
                levels_append(
                    dict_levels_get(match.group(1), debug_level_none))
                offsets_append(offset)
                last_line = line
            else:
                pos = find_insert_position(line)
                levels.insert(
                    pos, dict_levels_get(match.group(1), debug_level_none))
                offsets.insert(pos, offset)

        self.have_load_finished()
        yield False


class LogLine (list):

    _line_regex = default_log_line_regex()

    @classmethod
    def parse_full(cls, line_string):
        match = cls._line_regex.match(line_string.decode('utf8', errors='replace'))
        if match is None:
            # raise ValueError ("not a valid log line (%r)" % (line_string,))
            groups = [0, 0, 0, 0, "", "", 0, "", "", 0]
            return cls(groups)

        line = cls(match.groups())
        # Timestamp.
        line[0] = parse_time(line[0])
        # PID.
        line[1] = int(line[1])
        # Thread.
        line[2] = int(line[2], 16)
        # Level (this is handled in LineCache).
        line[3] = 0
        # Line.
        line[6] = int(line[6])
        # Message start offset.
        line[9] = match.start(9 + 1)

        for col_id in (4,   # COL_CATEGORY
                       5,   # COL_FILENAME
                       7,   # COL_FUNCTION,
                       8,):  # COL_OBJECT
            line[col_id] = sys.intern(line[col_id] or "")

        return line


class LogLines (object):

    def __init__(self, fileobj, line_cache):

        self.__fileobj = fileobj
        self.__line_cache = line_cache

    def __len__(self):

        return len(self.__line_cache.offsets)

    def __getitem__(self, line_index):

        offset = self.__line_cache.offsets[line_index]
        self.__fileobj.seek(offset)
        line_string = self.__fileobj.readline()
        line = LogLine.parse_full(line_string)
        msg = line_string[line[-1]:]
        line[-1] = msg
        return line

    def __iter__(self):

        size = len(self)
        i = 0
        while i < size:
            yield self[i]
            i += 1


class LogFile (Producer):

    def __init__(self, filename, dispatcher):

        import mmap

        Producer.__init__(self)

        self.logger = logging.getLogger("logfile")

        self.path = os.path.normpath(os.path.abspath(filename))
        self.__real_fileobj = open(filename, "rb")
        self.fileobj = mmap.mmap(
            self.__real_fileobj.fileno(), 0, access=mmap.ACCESS_READ)
        self.line_cache = LineCache(self.fileobj, dispatcher)
        self.line_cache.consumers.append(self)

    def start_loading(self):

        self.logger.debug("starting load")
        self.line_cache.start_loading()

    def get_load_progress(self):

        return self.line_cache.get_progress()

    def handle_load_started(self):

        # Chain up to our consumers:
        self.have_load_started()

    def handle_load_finished(self):
        self.logger.debug("finish loading")
        self.lines = LogLines(self.fileobj, self.line_cache)

        # Chain up to our consumers:
        self.have_load_finished()



# -*- coding: utf-8; mode: python; -*-
#
#  GStreamer Debug Viewer - View and analyze GStreamer debug log files
#
#  Copyright (C) 2007 René Stadler <mail@renestadler.de>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 3 of the License, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  this program.  If not, see <http://www.gnu.org/licenses/>.

"""GStreamer Debug Viewer GUI module."""

from gi.repository import Gtk
from gi.repository import Gdk

from GstDebugViewer import Data


class Color (object):

    def __init__(self, hex_24):

        if hex_24.startswith("#"):
            s = hex_24[1:]
        else:
            s = hex_24

        self._fields = tuple((int(hs, 16) for hs in (s[:2], s[2:4], s[4:],)))

    def gdk_color(self):

        return Gdk.color_parse(self.hex_string())

    def hex_string(self):

        return "#%02x%02x%02x" % self._fields

    def float_tuple(self):

        return tuple((float(x) / 255 for x in self._fields))

    def byte_tuple(self):

        return self._fields

    def short_tuple(self):

        return tuple((x << 8 for x in self._fields))


class ColorPalette (object):

    @classmethod
    def get(cls):

        try:
            return cls._instance
        except AttributeError:
            cls._instance = cls()
            return cls._instance


class TangoPalette (ColorPalette):

    def __init__(self):

        for name, r, g, b in [("black", 0, 0, 0,),
                              ("white", 255, 255, 255,),
                              ("butter1", 252, 233, 79),
                              ("butter2", 237, 212, 0),
                              ("butter3", 196, 160, 0),
                              ("chameleon1", 138, 226, 52),
                              ("chameleon2", 115, 210, 22),
                              ("chameleon3", 78, 154, 6),
                              ("orange1", 252, 175, 62),
                              ("orange2", 245, 121, 0),
                              ("orange3", 206, 92, 0),
                              ("skyblue1", 114, 159, 207),
                              ("skyblue2", 52, 101, 164),
                              ("skyblue3", 32, 74, 135),
                              ("plum1", 173, 127, 168),
                              ("plum2", 117, 80, 123),
                              ("plum3", 92, 53, 102),
                              ("chocolate1", 233, 185, 110),
                              ("chocolate2", 193, 125, 17),
                              ("chocolate3", 143, 89, 2),
                              ("scarletred1", 239, 41, 41),
                              ("scarletred2", 204, 0, 0),
                              ("scarletred3", 164, 0, 0),
                              ("aluminium1", 238, 238, 236),
                              ("aluminium2", 211, 215, 207),
                              ("aluminium3", 186, 189, 182),
                              ("aluminium4", 136, 138, 133),
                              ("aluminium5", 85, 87, 83),
                              ("aluminium6", 46, 52, 54)]:
            setattr(self, name, Color("%02x%02x%02x" % (r, g, b,)))


class ColorTheme (object):

    def __init__(self):

        self.colors = {}

    def add_color(self, key, *colors):

        self.colors[key] = colors


class LevelColorTheme (ColorTheme):

    pass


class LevelColorThemeTango (LevelColorTheme):

    def __init__(self):

        LevelColorTheme.__init__(self)

        p = TangoPalette.get()
        self.add_color(Data.debug_level_none, None, None, None)
        self.add_color(Data.debug_level_trace, p.black, p.aluminium2)
        self.add_color(Data.debug_level_fixme, p.black, p.butter3)
        self.add_color(Data.debug_level_log, p.black, p.plum1)
        self.add_color(Data.debug_level_debug, p.black, p.skyblue1)
        self.add_color(Data.debug_level_info, p.black, p.chameleon1)
        self.add_color(Data.debug_level_warning, p.black, p.orange1)
        self.add_color(Data.debug_level_error, p.white, p.scarletred1)
        self.add_color(Data.debug_level_memdump, p.white, p.aluminium3)


class ThreadColorTheme (ColorTheme):

    pass


class ThreadColorThemeTango (ThreadColorTheme):

    def __init__(self):

        ThreadColorTheme.__init__(self)

        t = TangoPalette.get()
        for i, color in enumerate([t.butter2,
                                   t.orange2,
                                   t.chocolate3,
                                   t.chameleon2,
                                   t.skyblue1,
                                   t.plum1,
                                   t.scarletred1,
                                   t.aluminium6]):
            self.add_color(i, color)
